"""
This module defines the arm handler to govern this enhancer device
"""
# ============================ imports =========================================
import threading
import time
from dataclasses import dataclass
from enum import Enum

from ArmSDK.Arm_Lib import Arm_Device
from txp import settings
import logging

from txp.common.edge import SingletonMeta

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class Servo(Enum):
    def __new__(cls, *values):
        """This defines the cannonical value (int) for the enum, and allows for other
        values. Useful for human readable strings.

        Reference for this solution:
            https://stackoverflow.com/questions/43202777/get-enum-name-from-multiple-values-python
        """
        obj = object.__new__(cls)
        obj._value_ = values[0]
        for other_value in values[1:]:
            cls._value2member_map_[other_value] = obj
        obj._all_values = values
        # human readable value useful for sending payload information to other systems
        obj.human_readable_value = values[1]
        return obj

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    def __str__(self):
        return self.human_readable_value

    NORMAL = 0, 'Normal', 'normal'
    THERMOGRAPHIC = 1, 'Thermographic', 'thermographic'

@dataclass
class ArmPositionDescriptor:
    servo_1:int=90 #base servo
    servo_2:int=90#firt linear extender servo
    servo_3:int=90 #second linea extender servo
    servo_4:int=90 #third linear extender servo
    servo_5:int=90 #cameras rotating enhancer servo
    servo_6 = 15 #dummy value to use servo position

    def get_positions_tuple(self):
        return (self.servo_1,self.servo_2,self.servo_3,self.servo_4,self.servo_5,self.servo_6)

    def get_positions_array(self):
        return list(self.get_positions_tuple())


class ServoConnectionStatusDescriptor:
    def __init__(self,servo_name,servo_index,status):
        self._servo_name = servo_name
        self._servo_index = servo_index
        self._status = status

    def set_servo_status(self,status):
        self._status = status

    def get_servo_status(self):
        return self._status


@dataclass
class ArmServoConnectionStatus:
    """
    ArmServoConnectionStatus validates the connection status of each servo
    if there is no connection with a specific servo this will set to false the status of the specific servo disconnection
    """
    servo_1_connection_status:ServoConnectionStatusDescriptor=ServoConnectionStatusDescriptor("servo_1",1,True)
    servo_2_connection_status:ServoConnectionStatusDescriptor=ServoConnectionStatusDescriptor("servo_2",2,True)
    servo_3_connection_status:ServoConnectionStatusDescriptor=ServoConnectionStatusDescriptor("servo_3",3,True)
    servo_4_connection_status:ServoConnectionStatusDescriptor=ServoConnectionStatusDescriptor("servo_4",4,True)
    servo_5_connection_status:ServoConnectionStatusDescriptor=ServoConnectionStatusDescriptor("servo_5",5,True)

    def set_servo_status(self,servo,status):
        attr = self.__getattribute__("servo_{}_connection_status".format(servo))
        attr.set_servo_status(status)

    def all_servos_connected(self):
        for index in range(1,6):
            if not self.__getattribute__("servo_{}_connection_status".format(index)).get_servo_status():
                return False
        return True

class ArmHandler(metaclass=SingletonMeta):

    def _get_arm_resource(self):
        log.debug("acquiring arm resource")
        return Arm_Device()

    def disable_torque(self):
        """
        Disable torque of the arm
        """
        arm = self._get_arm_resource()
        try:
            arm.Arm_serial_set_torque(0)
            time.sleep(0.1)
        finally:
            self._release_arm_resource(arm)


    def enable_torque(self):
        """
        Disable torque of the arm q
        """
        arm = self._get_arm_resource()
        try:
            arm.Arm_serial_set_torque(1)
            time.sleep(0.1)
        except Exception as ex:
            print(ex)
        finally:
            self._release_arm_resource(arm)

    def flush(self):
        """
        Flush arm commands
        """
        arm = self._get_arm_resource()
        try:
            arm.Arm_reset()
            time.sleep(0.01)
        finally:
            self._release_arm_resource(arm)

    def _get_servo_position(self,id):
        """
        get servo position by ID
        :param id: it is the number of the servo being queried its position
        :return: interger with angle value 0-180 degrees
        """

        arm = self._get_arm_resource()
        pos = None
        try:
            for retries in range(10):
                pos = arm.Arm_serial_servo_read(id)
                if pos is not None:
                    break
                else:
                    arm.Arm_reset()
                    time.sleep(0.01)
        finally:
            self._release_arm_resource(arm)
            return pos

    def get_positions(self):
        """
        returns the current position of all servos in a position object
        :return: ArmPositionDescriptor with arm position.
        """
        angle = ArmPositionDescriptor()
        angle.servo_1 = self._get_servo_position(1)
        angle.servo_2 = self._get_servo_position(2)
        angle.servo_3 = self._get_servo_position(3)
        angle.servo_4 = self._get_servo_position(4)
        angle.servo_5 = self._get_servo_position(5)
        log.info("position captured {}".format(angle))
        return angle

    def set_arm_angle_position(self, angle):
        arm = self._get_arm_resource()
        try:

            time_to_move = self._calculate_angle_move_time(arm,angle)
            log.debug("arm moving position")
            arm.Arm_serial_servo_write6(*angle,time_to_move)
            time.sleep(time_to_move/1000)
            log.debug("arm in requested position")
        except Exception as ex:
            log.error("error moving angle")
        finally:
            self._release_arm_resource(arm)

    def _release_arm_resource(self, arm):
        log.debug("releasing arm resource")
        del arm

    def arm_connection_status(self):
        """
        This retrieves the status of servos connected.
        :return: ArmServoConnectionStatus with statuses of all servos
        """
        arm = self._get_arm_resource()
        status = True
        try:
            for servo in range(1,6):
                log.debug("ping on servo {} connection".format(servo))
                ping_result = arm.Arm_ping_servo(id)
                if ping_result is None:
                    status = False
                log.debug("ping on servo {} connection result is {}".format(servo,status))
        finally:
            self._release_arm_resource(arm)
            return status

    def is_arm_servos_connected(self):
        return

    def execute_arm_action(self,angle):
        arm = self._get_arm_resource()
        time = self._calculate_angle_move_time(arm,angle)
        arm.Arm_serial_servo_write6(*angle,time)

    def _calculate_angle_move_time(self,arm,angle):
        return 2000


from txp.web.core_components.app_profiles import AppProfile
from txp.web.core_components.main_view import MainView
import txp.cloud.controllers.firestore as firestore_controller
import logging
from typing import Dict
import streamlit as st
import pandas as pd
import requests
from fastapi import status
from txp.cloud.common.utils import get_all_model_registries
from txp.common.ml.models import ModelRegistry
from txp.ml.training_service.common import TrainCommand
from st_aggrid import AgGrid
from st_aggrid.grid_options_builder import GridOptionsBuilder


# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class ManagementModelsView(MainView):
    """The management models View"""

    def __init__(self):
        super(ManagementModelsView, self).__init__(component_key="management_models_view")
        self.data = None
        self.assets = None
        self.tasks = None
        self.current_asset = None
        self.current_task = None
        self.model_registries = None
        self.models = None

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.ANALYTICS

    @classmethod
    def get_display_name(cls) -> str:
        return "Administrar modelos"

    def get_data(self):
        if self.data is None:
            self.db = firestore_controller.get_authenticated_client(
                self.app_state.authentication_info.role_service_account_credentials
            )

            self.data = firestore_controller.pull_recent_project_data(self.db, self._tenant_id)
            self.assets = {d["machine_id"]: d for d in self.data["machines"]}
            self.tasks = self.get_tasks_per_machines()
            self.current_asset = list(self.assets.keys())[0]
            self.current_task = self.tasks[self.current_asset][0]
            self.model_registries = self.get_all_model_registries_per_task()
            self.models = self.model_registries[self.current_task]
            self.url = st.secrets["service_url"]
            self.db.close()

    def get_all_model_registries_per_task(self):
        all_models = {}
        for asset, tasks in self.tasks.items():
            data_models_per_task = {task: get_all_model_registries(
                    self.db,
                    self._tenant_id,
                    asset,
                    task,
                    "txp_models_registry",
                ) for task in tasks}
            for task, data_model_dicts in data_models_per_task.items():
                dataclass_model_objects = [ModelRegistry(**data_model_dict) for data_model_dict in data_model_dicts]
                all_models.update({task: dataclass_model_objects})
        return all_models

    def get_tasks_per_machines(self):
        tasks_per_machine = {}
        for machine, values in self.assets.items():
            tasks = []
            if type(values["tasks"]) == dict:
                for task in values["tasks"].values():
                    tasks.append(task["task_id"])
            else:
                for task in values["tasks"]:
                    tasks.append(task['task_id'])
            tasks_per_machine.update({machine: tasks})
        return tasks_per_machine

    def _build_submenu_state(self) -> Dict:
        pass

    def _build_content_state(self) -> Dict:
        pass

    def update_asset(self):
        machine = st.session_state[f"{self.__class__.__name__}_asset_selectbox"]
        self.current_asset = machine
        self.current_task = self.tasks[machine][0]
        self.models = self.model_registries[self.current_task]

    def update_task(self):
        task = st.session_state[f"{self.__class__.__name__}_task_selectbox"]
        self.current_task = task
        self.models = self.model_registries[task]

    def _render_submenu(self):

        st.selectbox(
            "Activo",
            options=self.assets.keys(),
            key=f"{self.__class__.__name__}_asset_selectbox",
            on_change=self.update_asset,
        )

        st.selectbox(
            'Tarea',
            options=self.tasks[st.session_state[f"{self.__class__.__name__}_asset_selectbox"]],
            key=f"{self.__class__.__name__}_task_selectbox",
            on_change=self.update_task,
        )

    def _preprocessing_model_data(self, models):
        columns = ['filepath', 'task_id', 'value', 'feedback', 'creation_date', 'publishment_date', 'deprecation_date']
        model_record_list = []
        for model in models:
            model_dict = {}
            model_data = model.dict()
            for k, v in model_data['metadata'].items():
                if k in columns:
                    model_dict.update({k:v})
            for k, v in model_data['state'].items():
                if k in columns:
                    model_dict.update({k:v})
            model_dict.update({'filepath': model_data['file_path']})
            model_record_list.append(model_dict)

        model_records_with_desired_order_keys = []
        for model_record in model_record_list:
            reorder_model_record = {k: model_record[k] for k in columns}
            model_records_with_desired_order_keys.append(reorder_model_record)
        return model_records_with_desired_order_keys

    def _display_models(self, models):
        preprocessed_df = self._preprocessing_model_data(models)

        df = pd.DataFrame(preprocessed_df)
        gb = GridOptionsBuilder.from_dataframe(df)
        gb.configure_default_column(min_column_width=100)
        gb.configure_pagination()
        options = gb.build()
        AgGrid(df, gridOptions=options)

    def _config_model(self):
        new_task = st.session_state[f"{self.__class__.__name__}_config_model_selectbox"]
        dataset_name = st.session_state[f"{self.__class__.__name__}_dataset_name_input"]
        versions = st.session_state[f"{self.__class__.__name__}_dataset_version"].replace(" ", "").split(",")
        train_command = TrainCommand(
            dataset_name=dataset_name,
            dataset_version=versions,
            tenant_id=self._tenant_id,
            machine_id=self.current_asset,
            task_id=new_task
        )

        headers = {"Authorization": f"Bearer {st.secrets['service_auth_token']}"}
        response = requests.post(
            f"{self.url}/training/publish", json=train_command.dict(),
            headers=headers)

        if response.status_code == status.HTTP_201_CREATED:
            st.success(
                f'El task {new_task} fue publicado satisfactoriamente.'
            )
        elif response.status_code == status.HTTP_412_PRECONDITION_FAILED:
            st.warning(
                f"El task no puede ser publicado en este momento: {response.json()['Error']}"
            )
        else:
            st.error(
                f'No se pudo publicar el task {new_task}.'
            )
            st.json(response.json())

        self.model_registries = self.get_all_model_registries_per_task()
        self.models = self.model_registries[self.current_task]

    def _css_button(self):
        m = st.markdown("""
        <style>
        div.stButton > button:first-child {
        color: #0000CD;
        box-sizing: 5%;
        height: 2em;
        width: 300px;
        font-size:20px;
        border: 2px solid;
        border-color: #D3D3D3;
        border-radius: 10px;
        box-shadow: 1px 1px 1px 1px
        rgba(0, 0, 0, 0.1);
        padding-top: 0px;
        padding-bottom: 0px;
        right: 0px;
        position: absolute;
        }
        </style>""", unsafe_allow_html=True)

    def _render_content(self):
        self.get_data()
        self._css_button()
        st.header('Registro de modelos: lista de modelos disponibles')
        self._display_models(self.models)

        st.subheader('Configurar nueva tarea')
        st.selectbox(
                'Modelos disponibles:',
                options=self.tasks[self.current_asset],
                key=f"{self.__class__.__name__}_config_model_selectbox",
            )
        st.text_input(
            label="Ingrese el nombre del dataset",
            help="El nombre del dataset acordado por el equipo de anotación",
            key=f"{self.__class__.__name__}_dataset_name_input",
        )
        st.text_input(
            "Inserte el nombre de la versión:",
            help="Si desea especificar más de una version, ingrese los enunciados separados por coma.",
            key=f"{self.__class__.__name__}_dataset_version",
        )
        st.button('Actualizar', on_click=self._config_model)



import streamlit as st
import pandas as pd
import logging
import requests
import pytz
import datetime
import json
from fastapi import status
from txp.common.config import settings
from txp.common.ml.annotation import AnnotationLabelPayload

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class AnnotationWidget:
    def __init__(self, labels, machine_devices, data, dataset, client):
        self.machine_devices = machine_devices
        self.data = data
        self.dataset = dataset
        self.client = client

        self.labels = labels
        self.url = st.secrets['service_url']
        self._timestamps = []
        for sample in self.data:
            self._timestamps.append(self._change_utc_timezone(sample['observation_timestamp']))

    def tables_for_edge_type(self, logical_id):
        for device in self.machine_devices:
            if device['logical_id'] == logical_id:
                if device["device_kind"] == "Icomox":
                    return ['.time', '.fft', '.psd', '.time_metrics', '.fft_metrics', '.psd_metrics']
                else:
                    return ['.time']

    def _change_utc_timezone(self, timestamp):
        utc = pytz.timezone('UTC')
        timezone = pytz.timezone("America/Mexico_City")
        date_time = pd.to_datetime(timestamp)
        localized_timestamp = utc.localize(date_time)
        new_timezone = localized_timestamp.astimezone(timezone)
        strtime = new_timezone.strftime("%d/%m/%Y, %H:%M:%S,%f")
        return strtime

    def _render(self):
        """ Render visual components of annotation widget. """
        with st.form("annotation_form"):
            st.number_input(
                    'Número de versión:',
                    min_value=1,
                    step=1,
                    value=1,
                    key=f"{self.__class__.__name__}_actual_annotation_version"
            )

            st.markdown("Seleccione un rango de tiempo que desee adjudicarle una etiqueta.")
            st.selectbox(
                label="Desde:",
                options=['<Seleccionar>'] + sorted(set(self._timestamps)),
                key=f"{self.__class__.__name__}_annotation_since",
                index=0

            )
            st.selectbox(
                label="Hasta:",
                options=['<Seleccionar>'] + sorted(set(self._timestamps)),
                key=f"{self.__class__.__name__}_annotation_until",
                index=0
            )
            st.selectbox(
                label="Label",
                options=['<Seleccionar>'] + self.labels,
                key=f"{self.__class__.__name__}_annotation_label",
                index=0
            )
            st.form_submit_button("Anotar", on_click=self._annotate)

    def _annotate(self) -> None:
        """ Get annotation records. """
        if not st.session_state[f"{self.__class__.__name__}_annotation_since"] == '<Seleccionar>':
            version, samples, label = self._get_labeled_traces()

            # TODO: We should centralize the label handling in the core contacts
            json_label = {
                'label_value': label,
                'parameters': {}
            }

            annotation_records = []
            for sample in samples:
                annotation = AnnotationLabelPayload(
                    tenant_id=sample['tenant_id'],
                    dataset_name=f'{self.dataset}',
                    edge_logical_id=sample['edge_logical_id'],
                    observation_timestamp=sample['observation_timestamp'],
                    label=json_label,
                    version=version,
                )
                log.info(f"Generating annotation: {annotation.json()}")
                annotation_records.append(annotation.dict())
            try:
                headers = {"Authorization": f"Bearer {st.secrets['service_auth_token']}"}
                response = requests.post(f"{self.url}/annotation", json=annotation_records, headers=headers)
            except (ConnectionError, ConnectionRefusedError, Exception) as ce:
                st.error("No se pudo establecer con el servidor de anotación. "
                         "Contacte a soporte.")
            else:
                if response.status_code == status.HTTP_201_CREATED:
                    st.caption(f'{len(annotation_records)} muestras fueron etiquetadas como {label}')
                    print(response.json())
                else:
                    st.caption(f'{len(annotation_records)} no pudieron ser etiquetadas como {label}')
                    return None

    def _get_labeled_traces(self):
        """ selects the traces of figure that are within interval to annotate them. """

        version = st.session_state[f"{self.__class__.__name__}_actual_annotation_version"]

        since_timestamp = datetime.datetime.strptime(st.session_state[
            f"{self.__class__.__name__}_annotation_since"
        ], "%d/%m/%Y, %H:%M:%S,%f")
        until_timestamp = datetime.datetime.strptime(st.session_state[
            f"{self.__class__.__name__}_annotation_until"
        ], "%d/%m/%Y, %H:%M:%S,%f")
        label = st.session_state[
            f"{self.__class__.__name__}_annotation_label"
        ]

        labeled_samples = []
        for sample in self.data:
            mex_timestamp = self._change_utc_timezone(sample['observation_timestamp'])
            sample_timestamp = datetime.datetime.strptime(mex_timestamp, "%d/%m/%Y, %H:%M:%S,%f")
            if since_timestamp <= sample_timestamp <= until_timestamp:
                labeled_samples.append(sample)

        return version, labeled_samples, label




from SerialConnector import SerialConnector
from SmartMeshSDK.ApiDefinition   import ApiDefinition,    \
                                         IpMgrDefinition
from SmartMeshSDK.ApiException    import ConnectionError

class IpMgrConnectorSerialInternal(SerialConnector.SerialConnector):
    '''
    \ingroup ApiConnector
    
    \brief Internal class for IP manager connector, over Serial.
    '''

    HELLO_CMD      = (ApiDefinition.ApiDefinition.COMMAND, ['hello'])
    HELLO_RESP_CMD = (ApiDefinition.ApiDefinition.COMMAND, ['hello_response'])
    MGR_HELLO_CMD  = (ApiDefinition.ApiDefinition.NOTIFICATION, ['manager_hello'])

    def __init__(self, maxQSize=100):
        api_def = IpMgrDefinition.IpMgrDefinition()
        SerialConnector.SerialConnector.__init__(self, api_def, maxQSize)

        self.HELLO_IDS = {'hello':          self.api_def.nameToId(*self.HELLO_CMD),
                          'hello_response': self.api_def.nameToId(*self.HELLO_RESP_CMD),
                          'manager_hello':  self.api_def.nameToId(*self.MGR_HELLO_CMD),
                          }
        self.helloCmdIds = self.HELLO_IDS.values()
        self.isConnect = False
        self.TxPacketId = 0
        
    def connect(self, params):
        SerialConnector.SerialConnector.connect(self, params)
        try:
            self._sendHello()
        except:
            SerialConnector.SerialConnector.disconnect(self)
            raise

    def isHelloResponse(self, cmdId):
        return cmdId == self.HELLO_IDS['hello_response']
    
    #======================== TX ==============================================

    def _sendHello(self):
        'Send the client Hello to the manager'
        self.paramLock.acquire()
        self.isConnect = False
        cmdId,byteArray = self.api_def.serialize(['hello'],
                                                 {'version':  self.api_def.fieldOptions['protocolVersion'],
                                                  'cliSeqNo': (self.TxPacketId - 1) % 256,
                                                  'mode':     0,
                                                  })
        self.paramLock.release()
        resp = self._sendInternal(cmdId, False, byteArray)
        if resp['successCode'] == ApiDefinition.ApiDefinition.RC_OK:
            self.paramLock.acquire()
            self.RxPacketId = resp['mgrSeqNo']
            self.isConnect = True
            self.paramLock.release()
        else:
            self.disconnect()
    
    def _buildTxHeader(self,cmdId,isResponse,serializedFields):
        txHeader = []
        controlByte = 0
        if cmdId not in [self.HELLO_IDS['hello']]:
            controlByte |= 0x02
            if isResponse:
                controlByte |= 0x01
        
        txHeader.append(controlByte)
        txHeader.append(cmdId)
        self.paramLock.acquire()
        if isResponse:
            txHeader.append(self.RxPacketId)
        else:
            txHeader.append(self.TxPacketId)
        self.paramLock.release()
        txHeader.append(len(serializedFields))
        
        return txHeader
        
    def _ackIfNeeded(self,cmdId,isResponse):
        if  (
                (not isResponse) and
                (self.shouldAck)
            ):
            # send normal ACK (with RC=0)
            self._sendInternal(cmdId,True,[0])
            return True
        elif cmdId == self.HELLO_IDS['manager_hello'] and self.isConnect :
            raise ConnectionError("Unexpected manager_hello")
        return False
    
    def _iShouldAck(self,isResponse):
        return 
    
    #======================== RX ==============================================
    
    def _parseRxHeader(self,frameRx):
        
        controlByte     = frameRx[0]
        cmdId           = frameRx[1]
        packetId        = frameRx[2]
        length          = frameRx[3]
        payload         = frameRx[4:]
        
        # parse controlByte
        isResponse      = ((controlByte&0x01)>>0==1)
        self.shouldAck  = ((controlByte&0x02)>>1==1)
        
        return cmdId,length,isResponse,packetId,payload
    
    def isValidPacketId(self,cmdId,isResponse,packetId):
        self.paramLock.acquire()
        isValidId            = True
        isDuplicate          = False
        updateRxPacketId  = True
        if ((self.RxPacketId==None) or
            (cmdId in self.helloCmdIds)):
            isValidId             = True
            isDuplicate           = False
            updateRxPacketId   = True
        else:
            if isResponse:
                isValidId              = packetId==self.TxPacketId
                if isValidId :
                    self._incrementTxPacketId()
                isDuplicate            = False
                updateRxPacketId    = True
            else:
                if self.shouldAck:
                    # reliable transport
                    isValidId               = packetId==self.RxPacketId or packetId==(self.RxPacketId+1)%256
                    isDuplicate             = packetId==self.RxPacketId 
                    updateRxPacketId     = True
                else:
                    # unreliable transport
                    isValidId               = True
                    isDuplicate             = False
                    updateRxPacketId     = False
        self.paramLock.release()
        return (isValidId, isDuplicate, updateRxPacketId)
    
    #======================== packetId ========================================
    
    def _incrementTxPacketId(self):
        self.TxPacketId += 1
        if self.TxPacketId>0xff:
            self.TxPacketId=0



import json
from abc import ABC, abstractmethod
from typing import Dict, List

from google.cloud import storage, pubsub_v1
from google.oauth2 import service_account
import google.cloud.firestore as firestore
import pathlib
from txp.ml.prediction_service.service_manager.prediction_manager import Command
import logging
from joblib import dump
import os
from txp.cloud.common import utils as cu
from ray.actor import ActorHandle
from txp.common.ml.tasks import AssetTask
from txp.common.ml.models import ModelRegistry, ModelFeedback, ErrorModelFeedback
from txp.ml.training_service import training_exceptions as te


class BasicModelTrainer(ABC):
    """A trainer is a concrete implementation that knows how to
    train (encode) specific Tasks implementations."""

    def __init__(
            self,
            credentials_str,
            models_bucket_name,
            prediction_service_topic,
            models_registry_collection_name,
            log_level=logging.INFO,
    ):
        self.log_level = log_level
        logging.basicConfig(level=self.log_level)
        try:
            json_dict_service_account = json.loads(credentials_str, strict=False)
        except json.JSONDecodeError:
            logging.error("Error deconding JSON credentials value")
            raise ValueError("Error deconding JSON credentials value")

        try:
            self.credentials = service_account.Credentials.from_service_account_info(
                json_dict_service_account
            )
        except ValueError:
            logging.error("Error parsing Credentials, unexpected format.")

        self.models_registry_collection_name = models_registry_collection_name
        self.models_bucket_name = models_bucket_name
        self.prediction_service_topic = prediction_service_topic
        self.fetcher = None

    def get_label_value(self, task_definition: AssetTask, json_string):
        label_dict = json.loads(json_string)
        if "label_value" not in label_dict:
            raise te.WrongLabelValue("Empty label dict")
        if label_dict["label_value"] not in task_definition.label_def.labels:
            raise te.WrongLabelValue(label_dict["label_value"])
        return task_definition.label_def.label_to_int[label_dict["label_value"]]

    @abstractmethod
    def _train_task(
            self,
            task: ActorHandle,
            task_definition: AssetTask,
            dataset_name: str,
            dataset_versions: List,
            tenant_id: str,
            machine_id: str,
            task_id: str,
    ):
        """Train the Ray Actor reference that points to some concrete
        implementation of a txp ML Task class.

        Args:
            task (ActorHandle): object reference to the ray actor process.

            task_definition (Dict): A key value map with possible hyper parameters
                declared for the task.

            dataset_name: The dataset name in the models repository.

            dataset_versions: The list of labeling version that composes
                the dataset.

            tenant_id: The tenant ID to which this training belongs.

            machine_id: The asset to which this trained task belongs.

            task_id: The task identifier (task class name).
        """
        pass

    @abstractmethod
    def get_model_registry(self, task, tenant_id, machine_id, task_id, file_path) -> ModelRegistry:
        """This method returns model registry dataclass which contains all important information for a trained
        model
        """
        pass

    @abstractmethod
    def get_model_feedback(self, task) -> ModelFeedback:
        pass

    def train(
            self,
            task,
            task_definition: AssetTask,
            dataset_name,
            dataset_versions,
            tenant_id,
            machine_id,
            task_id,
    ):
        """Trains the Ray Actor Task and then stores the result model in
        GCS models storage.
        """
        gcs_task_path = cu.get_gcs_task_bucket_path(tenant_id, machine_id, task_id)
        gcs_file_path = f"{gcs_task_path}/{dataset_name}"
        file_name = cu.get_gcs_model_file_name(dataset_name, dataset_versions)
        model_registry_name = cu.get_gcs_dataset_prefix(dataset_name, dataset_versions)
        model_registry = self.get_model_registry(task, tenant_id, machine_id, task_definition.task_id,
                                                 f"{gcs_file_path}/{file_name}")
        firestore_db = firestore.Client(credentials=self.credentials, project=self.credentials.project_id)

        logging.info(f"Model created in firestore for tenant_id: {tenant_id}, machine_id: {machine_id}, "
                     f"task_id: {task_id}, dataset_name: {dataset_name}")
        cu.insert_ml_model(firestore_db, model_registry, model_registry_name,
                           self.models_registry_collection_name)

        logging.info(f"Starting training process for task "
                     f"identified with: {gcs_task_path}")

        error_feedback = None
        storage_task = None
        exception = None
        try:
            storage_task = self._train_task(
                task,
                task_definition,
                dataset_name,
                dataset_versions,
                tenant_id,
                machine_id,
                task_id,
            )
        except te.WrongLabelValue as e:
            error_feedback = ErrorModelFeedback(error_message=e.get_message())
            exception = e
        except te.NoParquetFileFound as e:
            error_feedback = ErrorModelFeedback(error_message=e.get_message(gcs_file_path=gcs_file_path))
            exception = e
        except te.ErrorWhileTrainingModel as e:
            error_feedback = ErrorModelFeedback(error_message=e.get_message())
            exception = e

        if error_feedback is not None:
            cu.set_error_ml_model(firestore_db, tenant_id, machine_id, task_id, model_registry_name,
                                  error_feedback, self.models_registry_collection_name)
            firestore_db.close()
            raise exception

        logging.info(f"Model identified with: {gcs_task_path} was successfully trained"
                     f" in memory")

        tmp_directory = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tmp")

        file_path = os.path.join(tmp_directory, file_name)
        pathlib.Path(tmp_directory).mkdir(parents=True, exist_ok=True)

        logging.info(f"Trying to store {gcs_task_path} as a local file in: {file_path}")

        dump(storage_task, file_path)

        logging.info(f"Model successfully dumped in local storage: {file_path}")

        storage_client = storage.Client(credentials=self.credentials, project=self.credentials.project_id)
        bucket = storage_client.bucket(self.models_bucket_name, user_project=self.credentials.project_id)
        logging.info(f"Trying to upload model to remote model storage: {gcs_file_path}")
        blob = bucket.blob(f"{gcs_file_path}/{file_name}")
        blob.upload_from_filename(file_path)
        storage_client.close()
        os.remove(file_path)
        feedback = self.get_model_feedback(task)
        cu.acknowledge_ml_model(firestore_db, tenant_id, machine_id, task_id, model_registry_name,
                                feedback, self.models_registry_collection_name)
        logging.info(f"Model registry acknowledged in firestore for tenant_id: {tenant_id}, machine_id: {machine_id}, "
                     f"task_id: {task_id}, dataset_name: {dataset_name}")

        firestore_db.close()
        logging.info(f"Model is already available in GCS at {self.models_bucket_name}/{gcs_file_path}")

    def send_command_to_prediction_service(self, tenant_id, asset_id, task_id):
        logging.basicConfig(level=self.log_level)
        logging.info(
            f"sending command to prediction service for tenant_id: {tenant_id}, asset_id: {asset_id}, "
            f"task_id: {task_id}"
        )
        publisher = pubsub_v1.PublisherClient(credentials=self.credentials)
        topic_path = publisher.topic_path(
            self.credentials.project_id, self.prediction_service_topic
        )
        future = publisher.publish(
            topic_path,
            data=json.dumps(
                {
                    "command": Command.UPDATE_TASK.value,
                    "parameters": {"tenant_id": tenant_id, "asset_id": asset_id, "task_id": task_id},
                }
            ).encode("utf-8"),
        )
        future.result()

    def publish_model(self, tenant_id, machine_id, task_id, dataset_name, dataset_versions):
        """Publish a model in response to an explicit /publish request by a client."""

        log_identifier_str = f"tenant_id: {tenant_id}, machine_id: {machine_id}, " \
                             f"task_id: {task_id}, dataset_name: {dataset_name}"
        firestore_db = firestore.Client(credentials=self.credentials, project=self.credentials.project_id)

        logging.info(f"Downgrading model in firestore for {log_identifier_str}")

        deprecation_res = cu.deprecate_active_ml_model(firestore_db, tenant_id, machine_id, task_id,
                                                       self.models_registry_collection_name)
        if deprecation_res:
            logging.info(f"Last active model was deprecated: {deprecation_res}")
        else:
            logging.info(f"There was not an active model for {log_identifier_str}")

        logging.info(f"Upgrading model in firestore for {log_identifier_str}")

        model_registry_name = cu.get_gcs_dataset_prefix(dataset_name, dataset_versions)
        publish_res = cu.publish_ml_model(firestore_db, tenant_id, machine_id, task_id, model_registry_name,
                                          self.models_registry_collection_name)
        if not publish_res:
            logging.info(f"Cannot Upgrade model in firestore for {log_identifier_str}, "
                         f"model does not exists or it is already active")
            firestore_db.close()
            return

        logging.info(f"Model upgraded in firestore for {log_identifier_str}")

        logging.info(f"Stating to publish model for {log_identifier_str}")

        file_name = cu.get_gcs_model_file_name(dataset_name, dataset_versions)
        gcs_task_path = cu.get_gcs_task_bucket_path(tenant_id, machine_id, task_id)
        gcs_file_path = f"{gcs_task_path}/{dataset_name}/{file_name}"
        firestore_db = firestore.Client(credentials=self.credentials, project=self.credentials.project_id)
        cu.update_model_in_task(firestore_db, tenant_id, machine_id, task_id, gcs_file_path)
        self.send_command_to_prediction_service(tenant_id, machine_id, task_id)
        firestore_db.close()

        logging.info(f"Model was published for {log_identifier_str}. ")
        logging.info(f"Model is already available at {self.models_bucket_name}/{gcs_file_path}")



import json
import os.path
import pathlib
import shutil
import google.api_core.exceptions
import pandas as pd
from google.oauth2 import service_account
from google.cloud import storage
from txp.cloud.common import utils as cu
import numpy as np
import logging
from txp.ml.training_service import training_exceptions as te


class BigqueryDatasetFetcher:
    def __init__(self, credentials_str, data_warehouse, log_level=logging.INFO):
        json_dict_service_account = json.loads(credentials_str, strict=False)
        self.credentials = service_account.Credentials.from_service_account_info(
            json_dict_service_account
        )
        self.data_warehouse = data_warehouse
        self.log_level = log_level

    def __get_dataframe_from_gcs(
            self, gcs_task_path, dataset_name, dataset_versions, table_name
    ):
        logging.basicConfig(level=self.log_level)
        storage_client = storage.Client(
            credentials=self.credentials, project=self.credentials.project_id
        )
        bucket = storage_client.bucket(
            self.data_warehouse, user_project=self.credentials.project_id
        )
        file_name = cu.get_gcs_dataset_file_name(
            gcs_task_path, dataset_name, table_name, dataset_versions
        )
        blob = bucket.blob(file_name)
        tmp_directory = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tmp")
        split_file_name = file_name.split("/")
        dataset_name = split_file_name[3]
        table_name = split_file_name[4]
        dataset_directory = os.path.join(tmp_directory, dataset_name)
        logging.info(f"Fetching remote dataset file for GCS path: {gcs_task_path}")
        if not os.path.isdir(dataset_directory):
            pathlib.Path(dataset_directory).mkdir(parents=True, exist_ok=True)
        file_path = os.path.join(dataset_directory, table_name)
        try:
            blob.download_to_filename(file_path)
        except google.api_core.exceptions.NotFound:
            logging.error(f"No such file found in GCS {file_name}")
            if os.path.exists(file_path):
                os.remove(file_path)
            return None
        df = pd.read_parquet(file_path)
        storage_client.close()
        shutil.rmtree(dataset_directory)
        return df

    def get_time_dataframe_from_gcs(
            self, dataset_name, dataset_versions, gcs_task_path
    ):
        df = self.__get_dataframe_from_gcs(
            gcs_task_path, dataset_name, dataset_versions, "time"
        )
        if df is None:
            raise te.NoParquetFileFound("time")
        return df

    def get_fft_dataframe_from_gcs(self, dataset_name, dataset_versions, gcs_task_path):
        df = self.__get_dataframe_from_gcs(
            gcs_task_path, dataset_name, dataset_versions, "fft"
        )
        if df is None:
            raise te.NoParquetFileFound("fft")
        df["fft"] = df["fft"].map(
            lambda fft: [
                [
                    np.complex128(complex(z["real"], z["imag"]))
                    for z in dimension["values"]
                ]
                for dimension in fft
            ]
        )
        return df

    def get_psd_dataframe_from_gcs(self, dataset_name, dataset_versions, gcs_task_path):
        df = self.__get_dataframe_from_gcs(
            gcs_task_path, dataset_name, dataset_versions, "psd"
        )
        if df is None:
            raise te.NoParquetFileFound("psd")
        df.rename(columns={"data": "psd"}, inplace=True)
        df["frequency"] = df["psd"].map(
            lambda data: [dimension["frequency"] for dimension in data]
        )
        df["psd"] = df["psd"].map(lambda data: [dimension["psd"] for dimension in data])
        return df

    def get_time_metrics_dataframe_from_gcs(
            self, dataset_name, dataset_versions, gcs_task_path
    ):
        df = self.__get_dataframe_from_gcs(
            gcs_task_path, dataset_name, dataset_versions, "time_metrics"
        )

        if df is None:
            raise te.NoParquetFileFound("time_metrics")
        return df

    def get_fft_metrics_dataframe_from_gcs(
            self, dataset_name, dataset_versions, gcs_task_path
    ):
        df = self.__get_dataframe_from_gcs(
            gcs_task_path, dataset_name, dataset_versions, "fft_metrics"
        )
        if df is None:
            raise te.NoParquetFileFound("fft_metrics")
        return df

    def get_psd_metrics_dataframe_from_gcs(
            self, dataset_name, dataset_versions, gcs_task_path
    ):
        df = self.__get_dataframe_from_gcs(
            gcs_task_path, dataset_name, dataset_versions, "psd_metrics"
        )
        if df is None:
            raise te.NoParquetFileFound("time_metrics")
        return df



import multiprocessing
import threading
import time
from typing import List, Tuple, Dict, Union, Optional

from txp.common.edge import DeviceId
from txp.devices.drivers.arm_robot.arm_handler import ArmHandler
from txp.devices.drivers.arm_robot.robot_driver import ArmRobotDriver
from txp.devices.drivers.arm_robot.thermal_robot import ThermalRobotDriver
from txp.devices.drivers.voyager.voyager_driver_host import VoyagerDriversHost
from txp.devices.drivers.icomox.icomox_driver_host import IcomoxDriversHost
from txp.devices.gateway_states_enums import DevicePairingStatus
from txp.devices.drivers.mock.mock_driver import MockDriversHost
from txp.devices.drivers.icomox.icomox_driver import IcomoxDriver
from txp.devices.drivers.voyager.voyager_driver import VoyagerDriver
from txp.devices.drivers.mock.mock_driver import MockDriver
from txp.devices.drivers.drivers_host import DriversHost
from txp.devices.drivers.arm_robot.robot_host import RobotDriversHost
from txp.common.edge.common import EdgeType, EdgeDescriptor, VirtualSensorObject, VirtualEdgeInformation
from txp.devices.drivers.usb.camera.camera_handler_base import CameraHandlerBase, CameraType
from txp.common.config import settings
import logging
import dataclasses
import copy

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

MAX_NUMBER_OF_DEVICES_ALLOWED_IN_PRODUCTION = 1

# Hardcoded virtual sensor object for the Thermocouple of a ThermoArm
_thermo_arm_thermocuple_sensor: VirtualSensorObject = VirtualSensorObject(
    "I2C-0x44", CameraType.THERMOCOUPLE
)


@dataclasses.dataclass
class _VirtualProductionRecord:
    """Private class to track the virtual production process information."""
    virtual_camera: VirtualSensorObject = VirtualSensorObject()
    virtual_thermocamera: VirtualSensorObject = VirtualSensorObject()
    virtual_positions_produced: List[Tuple[EdgeDescriptor, str]] = dataclasses.field(default_factory=list)
    torque_enabled: bool = True  # By default the torque will be enabled
    position_captured: Tuple[int] = dataclasses.field(default_factory=tuple)  # the last position captured
    image_captured: Optional = None  # The last image captured in the virtual production
    thermographic_img_captured: Optional = None  # The last thermographic image captured in the virtual production
    captured_positions: List[Dict] = dataclasses.field(default_factory=list)  # The list of captured positions

    def virtual_camera_produced(self) -> bool:
        return bool(self.virtual_camera)  # No data in virtual camera implies not produced

    def virtual_thermocamera_produced(self) -> bool:
        return bool(self.virtual_thermocamera)  # No data in virtual thermocamera implies not produced

    def copy(self) -> "_VirtualProductionRecord":
        return copy.deepcopy(self)


class ProductionManager:
    """
    TODO: this class will be part of the device manager and connection methods wont be here. This is subject to change
    the purpose of this module is to speed up production interface.
    """
    _device_to_host_class = {
        IcomoxDriver.device_name(): IcomoxDriversHost,
        VoyagerDriver.device_name(): VoyagerDriversHost,
        MockDriver.device_name(): MockDriversHost
    }

    def __init__(self, gateway_packages_queue, devices: List[str]):
        state_queue = multiprocessing.Manager().Queue()
        self._is_device_available = False
        self._driver_hosts_deprecated_list: List[DriversHost] = []
        self._driver_hosts: Dict[str, DriversHost] = {}
        log.info(f"Production manager configured devices: {devices}")
        for device in devices:
            if device == RobotDriversHost.get_driver_host_name() or \
                device == ThermalRobotDriver.device_name():  # RobotDriversHost not required in production.
                continue
            host_instance = self._device_to_host_class[device](gateway_packages_queue, state_queue)
            self._driver_hosts[device] = host_instance

        self._available_devices = {}
        self._virtual_production: _VirtualProductionRecord = None
        self._normal_camera_handler: CameraHandlerBase = None
        self._thermographic_camera_handler: CameraHandlerBase = None
        self._arm_handler = ArmHandler()

    def _get_indexed_host(self, host_name):
        host = self._driver_hosts.get(host_name, None)
        if not host:
            log.error(f"ProductionManager requested to search an edge of kind: {host_name}, but the hosts was "
                      f"never configured")
        return host

    def start_listening_for_an_edge(self, device_kind: str):
        drivers_host = self._get_indexed_host(device_kind)
        drivers_host.connect()

    def get_hosted_device(self) -> Tuple[str, DeviceId]:
        """

        :return: a tuple with host name with the DeviceId associated. if not devices registerd
        returns empty tuple
        """
        if len(self._available_devices) > 0:
            item = self._available_devices.popitem()
            host_associated = item[0]
            device_id_associated = (item[1])[0]
            return (host_associated, device_id_associated)
        else:
            tuple()

    def is_device_available(self):
        return self._is_device_available

    def wait_for_device_to_connect(self, device_kind: str, timeout_in_minutes=10):
        """
        blocks for timeout minutes until a device is connected or it time out
        returns true when 1 devices is connected to the nerwork

        :rtype: boolean
        :exception: if there are more than 1 device connected a InvalidNumberOfDevices exception is thrown.
        """

        self._clear_devices()
        self._hast_timed_out = False

        t = threading.Thread(target=self._wait_for_device_to_connect_thread_body,
                             args=[device_kind, timeout_in_minutes])
        t.start()

        while (not self._hast_timed_out) and (not self._is_device_available):
            pass

        devices_ids_associated = list(self._available_devices.values())

        if devices_ids_associated:
            if len(self._available_devices) > MAX_NUMBER_OF_DEVICES_ALLOWED_IN_PRODUCTION or \
                    len(devices_ids_associated[0]) > MAX_NUMBER_OF_DEVICES_ALLOWED_IN_PRODUCTION:
                raise InvalidNumberOfDevices("Invalid number of devices in production", devices_ids_associated)

        return self._is_device_available

    def _wait_for_device_to_connect_thread_body(self, device_kind, timeout_minutes):
        self._is_device_available = False
        elapsedTime = 0
        while not self._is_device_available:
            time.sleep(6)
            drivers_host = self._get_indexed_host(device_kind)
            if not drivers_host:
                return

            physical_ids = drivers_host.discover_physical_ids()
            if len(physical_ids) > 0:
                if device_kind in self._available_devices.keys():
                    self._available_devices[device_kind].extend(physical_ids)
                else:
                    self._available_devices[device_kind] = physical_ids
                log.info(f"Devices found in Production for: {device_kind}")

            if len(self._available_devices) > 0:
                self._is_device_available = True
                break

            elapsedTime += 1
            if elapsedTime == timeout_minutes:
                self._hast_timed_out = True
                log.info("timeout waiting for a device")
                break

        log.info("wait for devices to connect finished")

    def stop_listening_for_devices(self, device_kind: str):
        driver_host = self._get_indexed_host(device_kind)
        if not driver_host:
            return
        driver_host.disconnect()

    def _clear_devices(self):
        if len(self._available_devices):
            self._available_devices.clear()

    # ========================= Virtual Production Methods =======================
    """
    The methods below provides an API for use in the production of a virtual edge device. 
    Currently, the virtual edge is bound to the production of the Arm Robot device. 
    By design, the implementation tries to:
        - Detect normal camera
        - Detect thermocamera 
        - Capture virtual positions from the Arm robot.
    """

    def prepare_virtual_production(self, actuator_descriptor: EdgeDescriptor):
        """Initializes internal attributes to track the virtual production process
        of virtualized edges through the ARM ROBOT"""

        if actuator_descriptor.get_virtual_sensors() is None or not actuator_descriptor.get_virtual_sensors():
            self._virtual_production = _VirtualProductionRecord()
        else:
            virtual_camera_sensor = list(filter(
                lambda virtual_sensor: virtual_sensor.camera_type == CameraType.NORMAL,
                actuator_descriptor.get_virtual_sensors()
            ))[0]
            thermocamera_virtual_sensor = list(filter(
                lambda virtual_sensor: virtual_sensor.camera_type == CameraType.THERMOGRAPHIC,
                actuator_descriptor.get_virtual_sensors()
            ))[0]

            if virtual_camera_sensor is None or thermocamera_virtual_sensor is None:
                self._virtual_production = _VirtualProductionRecord()
                return

            self._virtual_production = _VirtualProductionRecord(
                virtual_camera_sensor,
                thermocamera_virtual_sensor
            )

    def get_next_virtual_production_step(self) -> DevicePairingStatus:
        """Returns the next appropriate Status in the virtual production"""
        if not self._virtual_production.virtual_camera_produced():
            return DevicePairingStatus.RequestVirtualCameraConnection

        elif not self._virtual_production.virtual_thermocamera_produced():
            return DevicePairingStatus.RequestVirtualThermocameraConnection

        else:  # Cameras already paired. Only Arm robot configuration required
            if not self._normal_camera_handler:
                log.info("Production Manager creating camera handler for Normal Camera")
                self._normal_camera_handler = CameraHandlerBase(
                    self._virtual_production.virtual_camera.camera_type.__str__(),
                    self._virtual_production.virtual_camera.physical_id
                )
                self._normal_camera_handler.connect()

            if not self._thermographic_camera_handler:
                log.info("Production Manager creating camera handler for Thermographic Camera")
                self._thermographic_camera_handler = CameraHandlerBase(
                    str(self._virtual_production.virtual_thermocamera.camera_type.__str__()),
                    self._virtual_production.virtual_thermocamera.physical_id
                )
                self._thermographic_camera_handler.connect()

            return DevicePairingStatus.VirtualRequestNextStepFromUser

    def detect_virtual_camera(self) -> Union[
        "DevicePairingStatus.VirtualCameraProducedSuccessfully",
        "DevicePairingStatus.VirtualCameraConnectionError"
    ]:
        # Precondition: The user clicked and confirmed that the camera was connected.
        # It should be found.
        camera_id = CameraHandlerBase.detect_usb_camera()
        if not camera_id:
            log.error("Trying to pair virtual camera, but not camera usb device was found.")
            return DevicePairingStatus.VirtualCameraConnectionError

        self._virtual_production.virtual_camera = VirtualSensorObject(camera_id, CameraType.NORMAL)
        return DevicePairingStatus.VirtualCameraProducedSuccessfully

    def detect_virtual_thermocamera(self) -> Union[
        "DevicePairingStatus.VirtualThermocameraProducedSuccessfully",
        "DevicePairingStatus.VirtualCameraConnectionError"
    ]:
        # Precondition: The user clicked and confirmed that the camera was connected.
        # It should be found.
        camera_id = CameraHandlerBase.detect_usb_camera()
        if not camera_id:
            log.error("Trying to pair virtual thermographic camera, but not camera usb device was found.")
            return DevicePairingStatus.VirtualThermocameraConnectionError

        self._virtual_production.virtual_thermocamera = VirtualSensorObject(camera_id,
                                                                            CameraType.THERMOGRAPHIC)

        return DevicePairingStatus.VirtualThermocameraProducedSuccessfully

    def get_virtual_production_record(self) -> _VirtualProductionRecord:
        log.info("Creating copy of virtual production record")
        start_time = time.time()
        r = self._virtual_production.copy()
        end_time = time.time()
        log.info(f"Time Elapsed creating copy of virtual production record: {end_time-start_time}")
        return r

    def enable_torque(self) -> None:
        self._virtual_production.torque_enabled = True
        self._arm_handler.enable_torque()
        time.sleep(0.1)
        return

    def disable_torque(self) -> None:
        self._virtual_production.torque_enabled = False
        self._arm_handler.disable_torque()
        time.sleep(0.1)
        return

    def capture_virtual_pos(self) -> None:
        log.info("Capturing images from cameras / Converting to Bytes")
        start_time = time.time()
        self._virtual_production.image_captured = bytes(self._normal_camera_handler.get_image())
        self._virtual_production.thermographic_img_captured = bytes(self._thermographic_camera_handler \
                                                                    .get_image())
        end_time = time.time()
        log.info(f"Total time elapsed for capture and conversion:{end_time-start_time}")
        self._virtual_production.position_captured = self._arm_handler.get_positions().get_positions_tuple()

    def reject_captured_pos(self) -> None:
        self._virtual_production.position_captured = tuple()
        self._virtual_production.image_captured = None
        self._virtual_production.thermographic_img_captured = None

    def confirm_captured_pos(self, actuator_edge_descriptor: EdgeDescriptor, asset_associated: str) -> None:
        hashed_virtual_id: str = EdgeDescriptor.get_hashed_virtual_id(
            self._virtual_production.position_captured
        )

        virtual_logical_id: str = EdgeDescriptor.get_complete_virtual_id(
                actuator_edge_descriptor.logical_id, hashed_virtual_id
        )

        virtual_information = VirtualEdgeInformation(
            self._virtual_production.position_captured,
            hashed_virtual_id
        )

        virtual_descriptor = EdgeDescriptor(
            virtual_logical_id,
            actuator_edge_descriptor.device_kind,
            EdgeType.ON_OFF_DEVICE.__str__(),
            {},
            actuator_edge_descriptor.perceptions.copy()
        )

        virtual_descriptor.set_physical_id(actuator_edge_descriptor.logical_id)

        """This If statement assigns the right sensors depending on the driver type of the device.
        ArmRobotDriver -> Normal Camera and Thermocamera
        ThermalRobotDriver -> Normal Camera, Thermocamera and thermocouple
        """
        if virtual_descriptor.device_kind == ArmRobotDriver.device_name():
            virtual_descriptor.set_virtual_sensors([
                self._virtual_production.virtual_camera,
                self._virtual_production.virtual_thermocamera
            ])
        else:
            virtual_descriptor.set_virtual_sensors([
                self._virtual_production.virtual_camera,
                self._virtual_production.virtual_thermocamera,
                _thermo_arm_thermocuple_sensor
            ])

        virtual_descriptor.set_virtual_edge_information(virtual_information)

        self._virtual_production.virtual_positions_produced.append((virtual_descriptor, asset_associated))
        self._virtual_production.position_captured = tuple()
        self._virtual_production.image_captured = None

    def confirm_all_captured_pos(self, actuator_edge_descriptor: EdgeDescriptor) -> List[EdgeDescriptor]:
        actuator_edge_descriptor.set_physical_id(actuator_edge_descriptor.logical_id)
        if actuator_edge_descriptor.device_kind == ArmRobotDriver.device_name():
            actuator_edge_descriptor.set_virtual_sensors([
                self._virtual_production.virtual_camera,
                self._virtual_production.virtual_thermocamera
            ])
        else:
            actuator_edge_descriptor.set_virtual_sensors([
                self._virtual_production.virtual_camera,
                self._virtual_production.virtual_thermocamera,
                _thermo_arm_thermocuple_sensor
            ])
        log.debug("Sensors added to Actuator Edge Descriptor")
        return self._virtual_production.virtual_positions_produced

    def finish_virtual_production(self):
        self._normal_camera_handler.disconnect()
        self._thermographic_camera_handler.disconnect()
        self._arm_handler.enable_torque()
        self._normal_camera_handler = None
        self._thermographic_camera_handler = None
        self._virtual_production = None


class InvalidNumberOfDevices(Exception):
    """
    Exception raised when there are more than one devices connected to the network
    when in production operation mode

    """
    pass


class NotValidDeviceRegisred(Exception):
    """
    Exception raised when get device method is called to report error since there is no device available.

    """
    pass



import numpy as np
from sklearn.tree import DecisionTreeClassifier
from sklearn import metrics
from sklearn.model_selection import train_test_split
import xgboost as xgb
from txp.ml.common.tasks.policy import Policy
from txp.ml.common.tasks.task import Task, Encoder
import pandas as pd


class VibrationBasicTask(Task):

    def __init__(self, task_definition=None, credentials_str=None, decision_tree_policy=False):

        schema = None
        task_file = None
        if task_definition is not None:
            schema = task_definition.schema_
            task_file = task_definition.task_data

        policy = VibrationDecisionTreePolicy()
        if not decision_tree_policy:
            parameters = None
            if task_definition is not None:
                parameters = task_definition.parameters
            policy = VibrationGradientBoostPolicy(parameters)

        super().__init__(VibrationEncoder(schema), policy, task_definition, credentials_str)

        if task_file is not None and credentials_str is not None:
            self.load_from_bucket(task_file)

            if schema is not None:
                self.encoder.set_schema(schema)

        self.ready = True
        print("VibrationBasicTask is ready.")

    def predict(self, sampling_window_rows) -> str:
        """
            Args:
                sampling_window_rows: filled firestore dictionary, this dict contains all the rows present
                                   in bigquery tables.
            Returns:
                Predicted VibrationBasicEvent.
        """
        print(f"VibrationBasicTask prediction request.")
        prediction_dataset = self.encoder.transform_input(sampling_window_rows)
        return self.task_definition.label_def.int_to_label[int(self.policy.predict(prediction_dataset)[0])]

    def train(self, dataset):
        self.policy.train(dataset, self.encoder.target)


class VibrationEncoder(Encoder):

    def __init__(self, schema):
        super().__init__(schema)

    def transform_input(self, signals_completed) -> pd.DataFrame:
        dataframe_row = {}
        for edge_logical_id in signals_completed:
            for perception_name in signals_completed[edge_logical_id]:
                for table_id in signals_completed[edge_logical_id][perception_name]:
                    rows = signals_completed[edge_logical_id][perception_name][table_id]
                    if table_id in ["time_metrics", "fft_metrics", "psd_metrics"]:
                        for row in rows:
                            for metric in self.schema[edge_logical_id][perception_name][table_id]:
                                dataframe_row[
                                    f'{metric}_{table_id}_{perception_name}_{edge_logical_id}_{row["dimension"]}'] \
                                    = row[metric]
                    else:
                        row = rows[0]
                        if table_id == "fft":
                            row["fft"] = [[np.complex128(complex(z["real"], z["imag"]))
                                           for z in dimension["values"]] for dimension in row["fft"]]
                        else:
                            row["psd"] = [dimension["psd"] for dimension in row["data"]]
                        for dimension in range(len(row[table_id])):
                            for i in range(len(row[table_id][dimension])):
                                dataframe_row[f"{table_id}_{perception_name}_{edge_logical_id}_{dimension}_{i}"] = \
                                    row[table_id][dimension][i].real

        dataset = pd.DataFrame()
        dataset = dataset.append(dataframe_row, ignore_index=True)
        columns = list(self.dataset.columns)
        columns.remove(self.target)

        return dataset[columns]

    def build_training_dataset(self, tables, target):
        self.target = target
        groups_time_df = tables['time_df'].groupby("observation_timestamp")
        groups_fft_df = tables['fft_df'].groupby(["observation_timestamp", "edge_logical_id", "perception_name"])
        groups_psd_df = tables['psd_df'].groupby(["observation_timestamp", "edge_logical_id", "perception_name"])
        groups_time_metrics_df = tables['time_metrics_df'].groupby(
            ["observation_timestamp", "edge_logical_id", "perception_name"])
        groups_fft_metrics_df = tables['fft_metrics_df'].groupby(["observation_timestamp", "edge_logical_id",
                                                                  "perception_name"])
        groups_psd_metrics_df = tables['psd_metrics_df'].groupby(["observation_timestamp", "edge_logical_id",
                                                                  "perception_name"])
        dataset = pd.DataFrame()

        for observation_timestamp, group_time_df in groups_time_df:
            row = self.__get_row(groups_time_metrics_df, groups_psd_metrics_df, groups_fft_metrics_df,
                                 observation_timestamp, groups_fft_df, groups_psd_df)
            if row:
                row[target] = group_time_df.iloc[0][target]
                dataset = dataset.append(row, ignore_index=True)

        self.dataset = dataset.replace(np.nan, 0)

        return self.dataset

    @staticmethod
    def __get_metrics(observation_timestamp, perception, group_dataframe, name, df_metrics, edge_logical_id):

        if (observation_timestamp, edge_logical_id, perception) not in group_dataframe.groups:
            return {}
        window_df = group_dataframe.get_group((observation_timestamp, edge_logical_id, perception))
        if len(window_df) < 3:
            return {}
        res = {}
        for _, signal in window_df.iterrows():
            for metric in df_metrics:
                res[f'{metric}_{name}_{perception}_{edge_logical_id}_{signal["dimension"]}'] = signal[metric]
        return res

    @staticmethod
    def __get_vector_magnitude(observation_timestamp, perception, group_dataframe, name, edge_logical_id):
        if (observation_timestamp, edge_logical_id, perception) not in group_dataframe.groups:
            return {}
        window_df = group_dataframe.get_group((observation_timestamp, edge_logical_id, perception))
        if len(window_df) != 1:
            return {}
        res = {}
        signal = window_df.iloc[0]
        for dimension in range(len(signal[name])):
            for i in range(len(signal[name][dimension])):
                res[f"{name}_{perception}_{edge_logical_id}_{dimension}_{i}"] = signal[name][dimension][i].real
        return res

    def __get_row(self, groups_time_metrics_df, groups_psd_metrics_df, groups_fft_metrics_df, observation_timestamp,
                  groups_fft_df, groups_psd_df):
        row = {}
        for edge_logical_id in self.schema:
            for perception in self.schema[edge_logical_id]:

                if "time_metrics" in self.schema[edge_logical_id][perception]:
                    row_metrics = self.__get_metrics(observation_timestamp, perception, groups_time_metrics_df,
                                                     "time_metrics",
                                                     self.schema[edge_logical_id][perception]["time_metrics"],
                                                     edge_logical_id)
                    if not row_metrics:
                        return {}
                    row = {**row, **row_metrics}

                if "psd_metrics" in self.schema[edge_logical_id][perception]:
                    row_metrics = self.__get_metrics(observation_timestamp, perception, groups_psd_metrics_df,
                                                     "psd_metrics",
                                                     self.schema[edge_logical_id][perception]["psd_metrics"],
                                                     edge_logical_id)
                    if not row_metrics:
                        return {}
                    row = {**row, **row_metrics}

                if "fft_metrics" in self.schema[edge_logical_id][perception]:
                    row_metrics = self.__get_metrics(observation_timestamp, perception, groups_fft_metrics_df,
                                                     "fft_metrics",
                                                     self.schema[edge_logical_id][perception]["fft_metrics"],
                                                     edge_logical_id)
                    if not row_metrics:
                        return {}
                    row = {**row, **row_metrics}

                if "fft" in self.schema[edge_logical_id][perception] and \
                        self.schema[edge_logical_id][perception]["fft"]:
                    vector_magnitude = self.__get_vector_magnitude(observation_timestamp, perception, groups_fft_df,
                                                                   "fft", edge_logical_id)
                    if not vector_magnitude:
                        return {}
                    row = {**row, **vector_magnitude}

                if "psd" in self.schema[edge_logical_id][perception] and \
                        self.schema[edge_logical_id][perception]["psd"]:
                    vector_magnitude = self.__get_vector_magnitude(observation_timestamp, perception, groups_psd_df,
                                                                   "psd", edge_logical_id)
                    if not vector_magnitude:
                        return {}
                    row = {**row, **vector_magnitude}

        return row


class VibrationDecisionTreePolicy(Policy):

    def __init__(self, policy_file: str = None):
        super().__init__(policy_file)
        if self.clf is None:
            self.clf = DecisionTreeClassifier()

    def name(self):
        return "VibrationDecisionTreePolicy"

    def train(self, dataset, target, test_size=0.3):
        features_col = list(dataset.columns)
        features_col.remove(target)
        x = dataset[features_col]
        y = dataset[target]
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=test_size, random_state=0)

        self.clf = self.clf.fit(x_train, y_train)
        y_prediction = self.clf.predict(x_test)

        self.accuracy = metrics.accuracy_score(y_test, y_prediction)
        self.trained = True

    def predict(self, tensor):
        if self.is_trained():
            return self.clf.predict(tensor)
        else:
            return None


class VibrationGradientBoostPolicy(Policy):

    def __init__(self, parameters=None):
        if parameters is None:
            self.param = {
                'max_depth': 1000,
                'eta': 1,
                'objective': 'multi:softmax',
                'num_class': 4,
                'num_round': 10
            }
        else:
            self.param = parameters
        super().__init__()

    def name(self):
        return "VibrationGradientBoostPolicy"

    def train(self, dataset, target, test_size=0.3):
        features_col = list(dataset.columns)
        features_col.remove(target)
        x = dataset[features_col]
        y = dataset[target]
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=test_size, random_state=0)

        train_matrix = xgb.DMatrix(x_train.values, y_train.values)
        test_matrix = xgb.DMatrix(x_test.values)

        params = self.param.copy()
        params.pop("num_round", None)

        self.clf = xgb.train(params, train_matrix, self.param["num_round"])
        y_prediction = self.clf.predict(test_matrix)

        self.accuracy = metrics.accuracy_score(y_test, y_prediction)
        self.trained = True

    def predict(self, tensor):
        if self.is_trained():
            matrix_to_predict = xgb.DMatrix(tensor.values)
            return self.clf.predict(matrix_to_predict)
        else:
            return None



#!/usr/bin/env python
import argparse
import logging
import os
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions, StandardOptions
from txp.cloud.pipelines import transformers as ts

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()
logger.setLevel(logging.INFO)

# Service account key path
os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "../credentials/pub_sub_to_bigquery_credentials.json"
INPUT_SUBSCRIPTION = "projects/PROJECT_ID/subscriptions/SUBSCRIPTION_NAME"
BIGQUERY_TABLE = "PROJECT_ID:DATASET_NAME.TABLE_NAME"
MODEL_SERVING_TOPIC_PREFIX = f'projects/{os.environ.get("GCP_PROJECT_ID", "tranxpert-mvp")}/topics/'
MODEL_SERVING_TOPIC_NAME = 'txp-model-serving-signals-test'


class FromProtoToJson(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element: bytes, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        import base64
        from txp.common.protos.gateway_package_pb2 import GatewayPackageProto
        from txp.cloud.common import utils

        proto_string = base64.b64decode(element)
        proto = GatewayPackageProto()
        proto.ParseFromString(proto_string)

        for e in utils.from_proto_to_json(proto):
            logging.info(f'arrived: {e["perception_name"]} - {e["edge_logical_id"]} - {e["tenant_id"]}')
            yield e


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input_subscription",
        help='Input PubSub subscription of the form "projects/<PROJECT>/subscriptions/<SUBSCRIPTION>."',
        default=INPUT_SUBSCRIPTION
    )
    parser.add_argument("--time_table", help="Output Time BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--fft_table", help="Output Fft BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--psd_table", help="Output Psd BigQuery Table", default=BIGQUERY_TABLE)

    parser.add_argument("--time_metrics_table", help="Output Time metrics BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--fft_metrics_table", help="Output Fft metrics BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--psd_metrics_table", help="Output Psd metrics BigQuery Table", default=BIGQUERY_TABLE)
    parser.add_argument("--model_signals_topic_name", help="topic for processing signals with ml task",
                        default=MODEL_SERVING_TOPIC_NAME)

    known_args, pipeline_args = parser.parse_known_args()
    model_serving_topic = MODEL_SERVING_TOPIC_PREFIX + known_args.model_signals_topic_name

    pipeline_options = PipelineOptions(pipeline_args)
    pipeline_options.view_as(StandardOptions).streaming = True

    with beam.Pipeline(options=pipeline_options) as p:
        signal_collection = (
            p
            | "ReadFromPubSub" >> beam.io.gcp.pubsub.ReadFromPubSub(
                subscription=known_args.input_subscription, timestamp_attribute=None
            )
            | "FromProtoToJson" >> beam.ParDo(FromProtoToJson())
        )
########################################################################################################################
        time_signal = (
            signal_collection
            | "TimeProcessing" >> beam.ParDo(ts.TimeProcessing())
        )
        (
            time_signal
            | "WriteTimeToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.time_table)
            | "WriteTimeToPubSub" >> beam.io.WriteToPubSub(topic=model_serving_topic, with_attributes=False)
        )
        (
            time_signal
            | "TimeMetrics" >> beam.ParDo(ts.TimeMetrics())
            | "WriteTimeMetricsToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.time_metrics_table)
            | "WriteTimeMetricsToPubSub" >> beam.io.WriteToPubSub(topic=model_serving_topic, with_attributes=False)
        )
########################################################################################################################
        fft_signal = (
            signal_collection
            | "FftProcessing" >> beam.ParDo(ts.FftProcessing())
        )
        (
            fft_signal
            | "WriteFftToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.fft_table)
            | "WriteFftToPubSub" >> beam.io.WriteToPubSub(topic=model_serving_topic, with_attributes=False)
        )
        (
            fft_signal
            | "FftMetrics" >> beam.ParDo(ts.FftMetrics())
            | "WriteFftMetricsToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.fft_metrics_table)
            | "WriteFftMetricsToPubSub" >> beam.io.WriteToPubSub(topic=model_serving_topic, with_attributes=False)
        )
########################################################################################################################
        psd_signal = (
            signal_collection
            | "PsdProcessing" >> beam.ParDo(ts.PsdProcessing())

        )
        (
            psd_signal
            | "WritePsdToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.psd_table)
            | "WritePsdToPubSub" >> beam.io.WriteToPubSub(topic=model_serving_topic, with_attributes=False)
        )
        (
            psd_signal
            | "PsdMetrics" >> beam.ParDo(ts.PsdMetrics())
            | "WritePsdMetricsToBigQuery" >> beam.ParDo(ts.WriteToBigQuery(), known_args.psd_metrics_table)
            | "WritePsdMetricsToPubSub" >> beam.io.WriteToPubSub(topic=model_serving_topic, with_attributes=False)
        )


if __name__ == "__main__":
    run()




import json
import time
from random import random
from google.auth import jwt
from google.cloud import pubsub_v1
from txp.common.configuration import SamplingWindow
from txp.common.edge import EdgeDescriptor, EdgeType
from txp.devices.drivers.mock.mock_driver import MockDriver, MockDriversHost
import base64
from multiprocessing.managers import BaseManager
from txp.devices.package_queue import PackageQueueProxy, PackageQueue
import multiprocessing
import datetime
from txp.common.configuration import SamplingParameters
from txp.devices.sampling_task import SamplingTask
from txp.cloud.common import utils as cu
import google.cloud.firestore as firestore
from google.oauth2 import service_account


CREDENTIALS_PATH = "../../../../common/credentials/pub_sub_to_bigquery_credentials.json"
PROJECT_ID = "tranxpert-mvp"
TOPIC_ID = "txp-telemetry-batch-test"
credentials = service_account.Credentials.from_service_account_file(CREDENTIALS_PATH)


class PubSubPublisher:
    def __init__(self, _credentials, topic_id):
        self.publisher = pubsub_v1.PublisherClient(credentials=_credentials)
        self.topic_path = self.publisher.topic_path(credentials.project_id, topic_id)

    def publish(self, data, gateway_id):
        result = self.publisher.publish(self.topic_path, data, gatewayId=gateway_id)
        return result


def main(edge, sampling_params, host, q_proxy):
    publisher = PubSubPublisher(credentials, TOPIC_ID)

    firestore_db = firestore.Client(credentials=credentials, project=credentials.project_id)
    config_id = \
        cu.pull_current_configuration_from_firestore(firestore_db, "labshowroom-001").to_dict()['configuration_id']
    firestore_db.close()

    task1 = SamplingTask(config_id, sampling_params, "labshowroom-001")

    host.receive_sampling_tasks(
        {
            edge.logical_id: task1
        }
    )
    host.start_sampling_windows()
    host.drivers[0].enable_observation_pulse()
    time.sleep(50)
    host.drivers[0].notify_sampling_window_completed()
    time.sleep(50)

    assert not q_proxy.is_empty()
    while not q_proxy.is_empty():
        package = q_proxy.get_package()
        proto = package.get_proto()
        encoded_payload = base64.b64encode(proto.SerializeToString())
        publisher.publish(encoded_payload, "test-mockup-gateway")
        time.sleep(random())


class CustomManager(BaseManager):
    pass


CustomManager.register("PackageQueue", PackageQueue, proxytype=PackageQueueProxy)


def edge_descriptors():
    mock_1_descriptor: EdgeDescriptor = EdgeDescriptor(
        "mock_1",
        MockDriver.device_name(),
        str(EdgeType.STREAM_ONLY_DEVICE),
        {
            MockDriver._PHYSICAL_ID_PARAM_FIELD: "001B638445E6",
            "clock_pulse": 15
        },
        {
            "VibrationMock": {
                "dimensions": [1, 1024],
                "sampling_frequency": 100000,
                "components_per_dimension": [
                    {
                        "components":
                            [
                                {
                                    "amplitude": 1,
                                    "frequency": 7.0,
                                    "phase": 0,
                                    "component_type": "sine"
                                },
                                {
                                    "amplitude": 1,
                                    "frequency": 13.0,
                                    "phase": 0,
                                    "component_type": "sine"
                                }
                            ],
                    }
                ],
                "mode": 3
            }
        },
    )
    return [mock_1_descriptor]


def sampling_parameters() -> SamplingParameters:
    params = SamplingParameters(
        SamplingWindow(10, 5),
        datetime.datetime.now().date(),
        datetime.datetime.now().date(),
        "1111111",
        datetime.datetime.now().time(),
        (datetime.datetime.now() + datetime.timedelta(seconds=15)).time(),
        []
    )
    return params


def states_queue():
    queue = multiprocessing.Manager().Queue()
    return queue


def queue_proxy() -> PackageQueueProxy:
    manager = CustomManager()
    manager.start()
    q_proxy: PackageQueueProxy = manager.PackageQueue()
    return q_proxy


def mock_host(q_proxy, states_q) -> MockDriversHost:
    h = MockDriversHost(
        q_proxy, states_q,
        **{"discovered_identifiers": ["001B638445E6", "001B638445E7", "001B638445E8"]}
    )
    h.connect()
    return h


if __name__ == "__main__":
    q_proxy = queue_proxy()
    states_q = states_queue()
    host = mock_host(q_proxy, states_q)
    host.add_drivers(edge_descriptors())
    main(edge_descriptors()[0], sampling_parameters(), host, q_proxy)

